﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Windows.Forms;

namespace AoDPatcher
{
	internal class PatchEngine
	{
		private readonly MainForm mainForm;

		private byte[] data;
		private long fileSize;

		private uint posPeHeader; // PEヘッダのファイル上の位置
		private uint countSections; // セクションの数
		private uint sizeCodeSection; // コードセクションのサイズ
		private uint posCodeSection; // コードセクションのファイル上の位置
		private uint addrBase; // ベースアドレス
		private uint alignSection; // セクションアラインメント
		private uint countImportDir; // インポートディレクトリの数
		private uint addrTextOutDc0; // TextOutDC0の埋め込みアドレス
		private uint addrGetTextWidth; // GetTextWidthの埋め込みアドレス
		private uint addrWinMmDll; // _INMM.DLLの埋め込みアドレス
		private uint posTextOutDc0; // TextOutDC0のファイル上の埋め込み位置
		private uint posGetTextWidth; // GetTextWidthのファイル上の埋め込み位置
		private uint posWinMmDll; // _INMM.DLLのファイル上の埋め込み位置
		private uint addrVarTextOutDc0Address; // varTextOutDC0Addressの割り当てアドレス
		private uint addrVarGetTextWidthAddress; // varGetTextWidthAddressの割り当てアドレス
		private uint rvaImportSection; // インポートセクションのRVA
		private uint posImportSection; // インポートセクションのファイル上の位置
		private uint addrGetModuleHandleA; // GetModuleHandleAの呼び出しアドレス
		private uint addrGetProcAddress; // GetProcAddressの呼び出しアドレス
		private uint posWinMmDllDef; // WINMM.DLLのファイル上の定義位置
		private uint posTextOutStart; // TextOutの処理開始位置
		private uint posTextOutEnd; // TextOutの処理後のジャンプ位置
		private uint posGetTextWidthStart; // GetTextWidthの処理開始位置
		private uint posGetTextWidthEnd; // GetTextWidthの処理後のジャンプ位置
		private uint posLatinToUpper; // Latin文字の大文字化処理の位置

		public PatchEngine(MainForm mainForm)
		{
			this.mainForm = mainForm;
		}

		public bool ExecPatch(string fileName)
		{
			LogWrite("Aod JP-Patcher Ver 0.04\n\n");

			try
			{
				ReadGameFile(fileName);

				if (!ParseHeader())
				{
					return false;
				}
				if (!ScanPatchLocation())
				{
					return false;
				}
				PatchBinary();
			}
			catch (Exception e)
			{
				MessageBox.Show(e.Message, "エラー");
				return false;
			}

			LogWrite("PatchBinary passed\n");

			return true;
		}

		public void SavePatchedFile(string fileName)
		{
			FileStream s = new FileStream(fileName, FileMode.Create, FileAccess.Write);
			s.Write(data, 0, data.Length);
			s.Close();
		}

		private void ReadGameFile(string fileName)
		{
			LogWrite("Phase 1 - ファイルの読み込み\n");

			FileInfo info = new FileInfo(fileName);
			fileSize = info.Length;

			data = new byte[fileSize];

			FileStream s = info.OpenRead();
			s.Read(data, 0, (int) fileSize);
			s.Close();

			LogWrite(string.Format("  TargetFile: {0}\n", fileName));
			LogWrite(string.Format("  FileSize: {0}Bytes\n", fileSize));

			LogWrite("Phase 1 passed\n\n");
		}

		private bool ParseHeader()
		{
			if (!ParseExeHeader())
			{
				return false;
			}
			if (!ParsePeHeader())
			{
				return false;
			}
			if (!ParseOptionHeader())
			{
				return false;
			}
			ParseSection();
			ParseImportTable();

			return true;
		}

		private bool ParseExeHeader()
		{
			LogWrite("Phase 2 - 実行ファイルの確認\n");

			if (GetWord(0) != 0x5A4D)
			{
				return false;
			}
			LogWrite("  exe識別子: $5A4D\n");

			posPeHeader = GetLong(60);
			LogWrite(string.Format("  PEの位置: {0:X8}\n", posPeHeader));

			LogWrite("Phase 2 passed\n\n");

			return true;
		}

		private bool ParsePeHeader()
		{
			LogWrite("Phase 3 - PEヘッダの確認\n");

			if (GetWord(posPeHeader) != 0x4550)
			{
				return false;
			}
			LogWrite("  PE識別子: $4550\n");

			countSections = GetWord(posPeHeader + 6);
			LogWrite(string.Format("  セクション数: {0}\n", countSections));

			uint posSymbolTable = GetLong(posPeHeader + 12);
			uint countSymbols = GetLong(posPeHeader + 16);
			LogWrite(string.Format("  シンボルテーブルの位置/数: ${0:X8}/${1:X8}\n", posSymbolTable, countSymbols));

			uint sizeOptionHeader = GetWord(posPeHeader + 20);
			LogWrite(string.Format("  オプションヘッダのサイズ: ${0:X4}\n", sizeOptionHeader));

			LogWrite("Phase 3 passed\n\n");

			return true;
		}

		private bool ParseOptionHeader()
		{
			LogWrite("Phase 4 - オプションヘッダの表示\n");

			uint offsetOptionHeader = posPeHeader + 24;
			if (GetWord(offsetOptionHeader) != 0x010B)
			{
				return false;
			}
			LogWrite("  オプション部識別子: $010B\n");

			sizeCodeSection = GetLong(offsetOptionHeader + 4);
			LogWrite(string.Format("  コードセクションのサイズ: ${0:X8} ({0}Bytes)\n", sizeCodeSection));

			uint sizeInitializedData = GetLong(offsetOptionHeader + 8);
			LogWrite(string.Format("  初期化済みデータセクションのサイズ: ${0:X8} ({0}Bytes)\n", sizeInitializedData));

			uint sizeUninitializedData = GetLong(offsetOptionHeader + 12);
			LogWrite(string.Format("  未初期化データセクションのサイズ: ${0:X8} ({0}Bytes)\n", sizeUninitializedData));

			uint addressEntryPoint = GetLong(offsetOptionHeader + 16);
			LogWrite(string.Format("  エントリーポイント: ${0:X8}\n", addressEntryPoint));

			posCodeSection = GetLong(offsetOptionHeader + 20);
			LogWrite(string.Format("  コードセクションの位置: ${0:X8}\n", posCodeSection));

			uint posDataSection = GetLong(offsetOptionHeader + 24);
			LogWrite(string.Format("  データセクションの位置: ${0:X8}\n", posDataSection));

			addrBase = GetLong(offsetOptionHeader + 28);
			LogWrite(string.Format("  ベースアドレス: ${0:X8}\n", addrBase));

			alignSection = GetLong(offsetOptionHeader + 32);
			LogWrite(string.Format("  セクション境界: ${0:X8} ({0}Bytes)\n", alignSection));

			uint alignmentFile = GetLong(offsetOptionHeader + 36);
			LogWrite(string.Format("  ファイル境界: ${0:X8} ({0}Bytes)\n", alignmentFile));

			uint sizeImage = GetLong(offsetOptionHeader + 56);
			LogWrite(string.Format("  イメージサイズ: ${0:X8}\n", sizeImage));

			uint sizeHeader = GetLong(offsetOptionHeader + 60);
			LogWrite(string.Format("  ヘッダサイズ: ${0:X8}\n", sizeHeader));

			uint rvaExportTable = GetLong(offsetOptionHeader + 96);
			uint sizeExportTable = GetLong(offsetOptionHeader + 100);
			LogWrite(string.Format("  Export Table address/size: ${0:X8}/${1:X8}\n", rvaExportTable, sizeExportTable));

			uint rvaImportTable = GetLong(offsetOptionHeader + 104);
			uint sizeImportTable = GetLong(offsetOptionHeader + 108);
			LogWrite(string.Format("  Import Table address/size: ${0:X8}/${1:X8}\n", rvaImportTable, sizeImportTable));

			uint rvaResourceTable = GetLong(offsetOptionHeader + 112);
			uint sizeResourceTable = GetLong(offsetOptionHeader + 116);
			LogWrite(string.Format("  Resource Table address/size: ${0:X8}/${1:X8}\n", rvaResourceTable,
			                       sizeResourceTable));

			uint rvaImportAddressTable = GetLong(offsetOptionHeader + 192);
			uint sizeImportAddressTable = GetLong(offsetOptionHeader + 196);
			countImportDir = sizeImportAddressTable/0x14;
			LogWrite(string.Format("  Import Address Table address/size: ${0:X8}/${1:X8}\n", rvaImportAddressTable,
			                       sizeImportAddressTable));

			LogWrite(string.Format("  * コード領域: ${0:X8}～${1:X8} (ファイル上: ${2:X8}～${3:X8})\n", addrBase + posCodeSection,
			                       addrBase + posCodeSection + sizeCodeSection - 1, posCodeSection,
			                       posCodeSection + sizeCodeSection - 1));
			LogWrite(string.Format("  * プログラムが開始されるアドレス: ${0:X8}\n", addrBase + addressEntryPoint));

			LogWrite("Phase 4 passed\n\n");

			return true;
		}

		private void ParseSection()
		{
			LogWrite("Phase 5 - セクションテーブルの表示");

			uint offsetSection = posPeHeader + 24 + 224;

			for (int i = 0; i < countSections; i++, offsetSection += 40)
			{
				LogWrite(string.Format("\n  セクション[{0}]\n", i));

				string nameSection = GetString(offsetSection, 8);
				LogWrite(string.Format("  セクション名: {0}\n", nameSection));

				uint virtualSize = GetLong(offsetSection + 8);
				LogWrite(string.Format("  セクション仮想サイズ: ${0:X8} ({0}Bytes)\n", virtualSize));

				uint virtualAddress = GetLong(offsetSection + 12);
				LogWrite(string.Format("  セクション仮想アドレス: ${0:X8}\n", virtualAddress));

				uint sizeRawData = GetLong(offsetSection + 16);
				LogWrite(string.Format("  セクション実サイズ: ${0:X8} ({0}Bytes)\n", sizeRawData));

				uint posRawData = GetLong(offsetSection + 20);
				LogWrite(string.Format("  セクション実位置: ${0:X8}\n", posRawData));

				if (nameSection.Equals(".rdata"))
				{
					uint addressRdataEnd = Ceiling(addrBase + virtualAddress + virtualSize, alignSection) - 1;
					LogWrite(string.Format("  * 未使用領域: ${0:X8}～${1:X8}\n", addrBase + virtualAddress + virtualSize + 1,
					                       addressRdataEnd));
					addrTextOutDc0 = addressRdataEnd - 0x0000001F;
					addrGetTextWidth = addressRdataEnd - 0x00000014;
					addrWinMmDll = addressRdataEnd - 0x0000002F;
					posTextOutDc0 = posRawData + sizeRawData - 0x00000020;
					posGetTextWidth = posRawData + sizeRawData - 0x00000015;
					posWinMmDll = posRawData + sizeRawData - 0x00000030;
				}
				else if (nameSection.Equals(".data"))
				{
					uint addressDataEnd = Ceiling(addrBase + virtualAddress + virtualSize, alignSection) - 1;
					LogWrite(string.Format("  * 未使用領域: ${0:X8}～${1:X8}\n", addrBase + virtualAddress + virtualSize + 1,
					                       addressDataEnd));
					addrVarTextOutDc0Address = addressDataEnd - 0x00000007;
					addrVarGetTextWidthAddress = addressDataEnd - 0x00000003;
				}
				else if (nameSection.Equals(".idata"))
				{
					rvaImportSection = virtualAddress;
					posImportSection = posRawData;
				}
			}

			LogWrite("Phase 5 passed\n\n");
		}

		private void ParseImportTable()
		{
			LogWrite("Phase 6 - インポートディレクトリテーブルの表示\n");
			LogWrite(string.Format("  インポートディレクトリテーブル数: {0}\n", countImportDir));

			uint offsetImportDir = posImportSection;
			uint rvaImportLookupTable = GetLong(offsetImportDir);
			while (rvaImportLookupTable != 0)
			{
				LogWrite(string.Format("\n  ImportLookupTableRVA: ${0:X8}\n", rvaImportLookupTable));

				uint timeStamp = GetLong(offsetImportDir + 4);
				LogWrite(string.Format("  TimeStamp: ${0:X8}\n", timeStamp));

				uint forwarderChain = GetLong(offsetImportDir + 8);
				LogWrite(string.Format("  FowarderChain: ${0:X8}\n", forwarderChain));

				uint rvaName = GetLong(offsetImportDir + 12);
				string name = GetString(posImportSection + rvaName - rvaImportSection, 0);
				LogWrite(string.Format("  NameRVA: ${0:X8} ({1})\n", rvaName, name));

				uint rvaImportAddressTable = GetLong(offsetImportDir + 16);
				LogWrite(string.Format("  ImportAddressTableRVA: ${0:X8}\n", rvaImportAddressTable));

				uint offsetImportLookupTable = posImportSection + rvaImportLookupTable - rvaImportSection;
				uint i = 0;
				uint entry = GetLong(offsetImportLookupTable);
				while (entry != 0)
				{
					uint hint;
					string importName;

					if ((entry & 0x80000000) != 0)
					{
						hint = entry & 0x7FFFFFFF;
						importName = "(Index)";
					}
					else
					{
						uint rvaHintNameTable = entry & 0x7FFFFFFF;
						uint offsetHintNameTable = posImportSection + rvaHintNameTable - rvaImportSection;
						hint = GetWord(offsetHintNameTable);
						importName = GetString(offsetHintNameTable + 2, 0);

						if (importName.Equals("GetModuleHandleA"))
						{
							addrGetModuleHandleA = addrBase + rvaImportAddressTable + i*4;
						}
						if (importName.Equals("GetProcAddress"))
						{
							addrGetProcAddress = addrBase + rvaImportAddressTable + i*4;
						}
					}
					LogWrite(string.Format("    {0:D4} {1}\n", hint, importName));

					offsetImportLookupTable += 4;

					i++;
					entry = GetLong(offsetImportLookupTable);
				}

				offsetImportDir += 0x14;

				rvaImportLookupTable = GetLong(offsetImportDir);
			}

			LogWrite("Phase 6 passed\n\n");
		}

		private bool ScanPatchLocation()
		{
			if (!ScanWinMmDll())
			{
				return false;
			}
			if (!ScanTextOutStart())
			{
				return false;
			}
			if (!ScanTextOutEnd())
			{
				return false;
			}
			if (!ScanGetTextWidthStart())
			{
				return false;
			}
			if (!ScanGetTextWidthEnd())
			{
				return false;
			}
			if (!ScanLatinToUpper())
			{
				return false;
			}
			return true;
		}

		private bool ScanWinMmDll()
		{
			LogWrite("ScanBinary - 特定バイナリを探す\n");
			LogWrite("  \"%06 WINMM.dll\"を探します\n");
			byte[] pattern = new byte[] {0x57, 0x49, 0x4E, 0x4D, 0x4D, 0x2E, 0x64, 0x6C, 0x6C};
			List<uint> l = BinaryScan(data, pattern, 0, (uint) fileSize);
			if (l.Count == 0)
			{
				return false;
			}
			posWinMmDllDef = l[0];
			LogWrite("ScanBinary passed\n\n");

			return true;
		}

		private bool ScanTextOutStart()
		{
			LogWrite("ScanBinary - 特定バイナリを探す\n");
			LogWrite("  \"%07 TextOutの処理を埋め込む位置\"を探します\n");
			byte[] pattern = new byte[]
			                 	{
			                 		0x8B, 0x45, 0x18, 0x8B, 0x48, 0x0C, 0x89, 0x4D, 0xC8, 0x8B, 0x55, 0x18, 0x8B, 0x42, 0x18, 0x89,
			                 		0x45, 0xA4, 0x8B, 0x45, 0xA4, 0x99
			                 	};
			List<uint> l = BinaryScan(data, pattern, posCodeSection, sizeCodeSection);
			if (l.Count == 0)
			{
				return false;
			}
			posTextOutStart = l[0];
			LogWrite("ScanBinary passed\n\n");

			return true;
		}

		private bool ScanTextOutEnd()
		{
			LogWrite("ScanBinary - 特定バイナリを探す\n");
			LogWrite("  \"%08 TextOutの処理が終了したらジャンプさせる位置\"を探します\n");
			byte[] pattern = new byte[]
			                 	{
			                 		0x8B, 0x4D, 0x18, 0x8B, 0x11, 0x8B, 0x4D, 0x18, 0xFF, 0x52, 0x0C, 0x8B, 0x45, 0x0C, 0x8B, 0xE5,
			                 		0x5D, 0xC2, 0x18, 0x00
			                 	};
			List<uint> l = BinaryScan(data, pattern, posCodeSection, sizeCodeSection);
			if (l.Count == 0)
			{
				return false;
			}
			posTextOutEnd = l[0];
			LogWrite("ScanBinary passed\n\n");

			return true;
		}

		private bool ScanGetTextWidthStart()
		{
			LogWrite("ScanBinary - 特定バイナリを探す\n");
			LogWrite("  \"%09 文字列幅の処理を埋め込む位置\"を探します\n");
			byte[] pattern = new byte[]
			                 	{
			                 		0xC7, 0x45, 0xFC, 0x00, 0x00, 0x00, 0x00, 0x8B, 0x45, 0x08, 0x8A, 0x08, 0x88, 0x4D, 0xF8, 0x8B,
			                 		0x55, 0xF8, 0x81, 0xE2, 0xFF, 0x00, 0x00, 0x00, 0x8B, 0x45, 0x08, 0x83, 0xC0, 0x01, 0x89, 0x45,
			                 		0x08, 0x85, 0xD2, 0x0F, 0x84
			                 	};
			List<uint> l = BinaryScan(data, pattern, posCodeSection, sizeCodeSection);
			if (l.Count == 0)
			{
				return false;
			}
			posGetTextWidthStart = l[0];
			LogWrite("ScanBinary passed\n\n");

			return true;
		}

		private bool ScanGetTextWidthEnd()
		{
			LogWrite("ScanBinary - 特定バイナリを探す\n");
			LogWrite("  \"%10 文字列幅の処理が終了したらジャンプさせる位置(の2バイト前)\"を探します\n");
			byte[] pattern = new byte[] {0x8B, 0x45, 0xFC, 0x8B, 0xE5, 0x5D, 0xC2, 0x04, 0x00};
			List<uint> l = BinaryScan(data, pattern, posCodeSection, sizeCodeSection);
			if (l.Count == 0)
			{
				return false;
			}
			posGetTextWidthEnd = l[0];
			LogWrite("ScanBinary passed\n\n");

			return true;
		}

		private bool ScanLatinToUpper()
		{
			LogWrite("ScanBinary - 特定バイナリを探す\n");
			LogWrite("  \"%XX Latin文字を大文字化している処理の位置\"を探します。\n");
			byte[] pattern = new byte[] {0x75, 0x18, 0x8B, 0xFB, 0x83, 0xC9, 0xFF, 0x33};
			List<uint> l = BinaryScan(data, pattern, posCodeSection, sizeCodeSection);
			if (l.Count == 0)
			{
				return false;
			}
			posLatinToUpper = l[0];
			LogWrite("ScanBinary passed\n\n");

			return true;
		}

		private void PatchBinary()
		{
			LogWrite("PatchBinary - パッチをあてる\n");
			PatchTextOut();
			PatchGetTextWidth();
			PatchWinMmDll();
			EmbedTextOutDc0();
			EmbedGetTextWidth();
			EmbedWinMmDll();
			PatchLatinToUpper();
		}

		private void PatchTextOut()
		{
			LogWrite("  proc TextOut書き換え\n");
			uint offset = posTextOutStart;
			PatchByte(data, offset, 0x83);
			offset++;
			PatchByte(data, offset, 0x3D);
			offset++;
			PatchLong(data, offset, addrVarTextOutDc0Address,
			          string.Format("%02 ${0:X8} varTextOutDC0Address", addrVarTextOutDc0Address));
			offset += 4;
			PatchByte(data, offset, 0x00);
			offset++;
			PatchByte(data, offset, 0x75);
			offset++;
			PatchByte(data, offset, 0x29);
			offset++;
			PatchByte(data, offset, 0x68);
			offset++;
//			PatchLong(data, offset, addressBase + posWinMmDll, string.Format("%06 ${0:X8} WINMM.dll", posWinMmDll));
			PatchLong(data, offset, addrWinMmDll, string.Format("%06 ${0:X8} WINMM.dll", addrWinMmDll));
			offset += 4;
			PatchByte(data, offset, 0xFF);
			offset++;
			PatchByte(data, offset, 0x15);
			offset++;
			PatchLong(data, offset, addrGetModuleHandleA,
			          string.Format("%04 ${0:X8} GetModuleHandleA", addrGetModuleHandleA));
			offset += 4;
			PatchByte(data, offset, 0x85);
			offset++;
			PatchByte(data, offset, 0xC0);
			offset++;
			PatchByte(data, offset, 0x75);
			offset++;
			PatchByte(data, offset, 0x01);
			offset++;
			PatchByte(data, offset, 0xF4);
			offset++;
			PatchByte(data, offset, 0x68);
			offset++;
			PatchLong(data, offset, addrTextOutDc0, string.Format("%00 ${0:X8} TextOutDC0", addrTextOutDc0));
			offset += 4;
			PatchByte(data, offset, 0x50);
			offset++;
			PatchByte(data, offset, 0xFF);
			offset++;
			PatchByte(data, offset, 0x15);
			offset++;
			PatchLong(data, offset, addrGetProcAddress, string.Format("%05 ${0:X8} GetProcAddress", addrGetProcAddress));
			offset += 4;
			PatchByte(data, offset, 0x85);
			offset++;
			PatchByte(data, offset, 0xC0);
			offset++;
			PatchByte(data, offset, 0x75);
			offset++;
			PatchByte(data, offset, 0x01);
			offset++;
			PatchByte(data, offset, 0xF4);
			offset++;
			PatchByte(data, offset, 0xA3);
			offset++;
			PatchLong(data, offset, addrVarTextOutDc0Address,
			          string.Format("%02 ${0:X8} varTextOutDC0Address", addrVarTextOutDc0Address));
			offset += 4;
			PatchByte(data, offset, 0x8B);
			offset++;
			PatchByte(data, offset, 0x4D);
			offset++;
			PatchByte(data, offset, 0x94);
			offset++;
			PatchByte(data, offset, 0x8B);
			offset++;
			PatchByte(data, offset, 0x51);
			offset++;
			PatchByte(data, offset, 0x0C);
			offset++;
			PatchByte(data, offset, 0x81);
			offset++;
			PatchByte(data, offset, 0x7A);
			offset++;
			PatchByte(data, offset, 0x19);
			offset++;
			PatchByte(data, offset, 0x62);
			offset++;
			PatchByte(data, offset, 0x6F);
			offset++;
			PatchByte(data, offset, 0x6C);
			offset++;
			PatchByte(data, offset, 0x64);
			offset++;
			PatchByte(data, offset, 0xB8);
			offset++;
			PatchByte(data, offset, 0x01);
			offset++;
			PatchByte(data, offset, 0x00);
			offset++;
			PatchByte(data, offset, 0x00);
			offset++;
			PatchByte(data, offset, 0x00);
			offset++;
			PatchByte(data, offset, 0x74);
			offset++;
			PatchByte(data, offset, 0x02);
			offset++;
			PatchByte(data, offset, 0x33);
			offset++;
			PatchByte(data, offset, 0xC0);
			offset++;
			PatchByte(data, offset, 0x8B);
			offset++;
			PatchByte(data, offset, 0xD0);
			offset++;
			PatchByte(data, offset, 0x80);
			offset++;
			PatchByte(data, offset, 0xB9);
			offset++;
			PatchByte(data, offset, 0x14);
			offset++;
			PatchByte(data, offset, 0x04);
			offset++;
			PatchByte(data, offset, 0x00);
			offset++;
			PatchByte(data, offset, 0x00);
			offset++;
			PatchByte(data, offset, 0x00);
			offset++;
			PatchByte(data, offset, 0x74);
			offset++;
			PatchByte(data, offset, 0x03);
			offset++;
			PatchByte(data, offset, 0x83);
			offset++;
			PatchByte(data, offset, 0xC8);
			offset++;
			PatchByte(data, offset, 0x02);
			offset++;
			PatchByte(data, offset, 0x50);
			offset++;
			PatchByte(data, offset, 0xFF);
			offset++;
			PatchByte(data, offset, 0x75);
			offset++;
			PatchByte(data, offset, 0x1C);
			offset++;
			PatchByte(data, offset, 0x8B);
			offset++;
			PatchByte(data, offset, 0x41);
			offset++;
			PatchByte(data, offset, 0x10);
			offset++;
			PatchByte(data, offset, 0x48);
			offset++;
			PatchByte(data, offset, 0x48);
			offset++;
			PatchByte(data, offset, 0x48);
			offset++;
			PatchByte(data, offset, 0x2B);
			offset++;
			PatchByte(data, offset, 0xC2);
			offset++;
			PatchByte(data, offset, 0x8B);
			offset++;
			PatchByte(data, offset, 0x51);
			offset++;
			PatchByte(data, offset, 0x0C);
			offset++;
			PatchByte(data, offset, 0x81);
			offset++;
			PatchByte(data, offset, 0x7A);
			offset++;
			PatchByte(data, offset, 0x0A);
			offset++;
			PatchByte(data, offset, 0x61);
			offset++;
			PatchByte(data, offset, 0x72);
			offset++;
			PatchByte(data, offset, 0x69);
			offset++;
			PatchByte(data, offset, 0x61);
			offset++;
			PatchByte(data, offset, 0x75);
			offset++;
			PatchByte(data, offset, 0x01);
			offset++;
			PatchByte(data, offset, 0x48);
			offset++;
			PatchByte(data, offset, 0x50);
			offset++;
			PatchByte(data, offset, 0x8B);
			offset++;
			PatchByte(data, offset, 0x55);
			offset++;
			PatchByte(data, offset, 0x18);
			offset++;
			PatchByte(data, offset, 0xFF);
			offset++;
			PatchByte(data, offset, 0x72);
			offset++;
			PatchByte(data, offset, 0x2C);
			offset++;
			PatchByte(data, offset, 0xFF);
			offset++;
			PatchByte(data, offset, 0x75);
			offset++;
			PatchByte(data, offset, 0x08);
			offset++;
			PatchByte(data, offset, 0xFF);
			offset++;
			PatchByte(data, offset, 0x75);
			offset++;
			PatchByte(data, offset, 0x10);
			offset++;
			PatchByte(data, offset, 0xFF);
			offset++;
			PatchByte(data, offset, 0x75);
			offset++;
			PatchByte(data, offset, 0x0C);
			offset++;
			PatchByte(data, offset, 0xFF);
			offset++;
			PatchByte(data, offset, 0x15);
			offset++;
			PatchLong(data, offset, addrVarTextOutDc0Address,
			          string.Format("%02 ${0:X8} varTextOutDC0Address", addrVarTextOutDc0Address));
			offset += 4;
			PatchByte(data, offset, 0xE9);
			offset++;
			PatchLong(data, offset, posTextOutEnd - offset - 4, string.Format("%08 ${0:X8} TextOutEnd", posTextOutEnd));
			offset += 4;
			PatchByte(data, offset, 0x90);
			offset++;
			PatchByte(data, offset, 0x90);
			LogWrite("\n");
		}

		private void PatchGetTextWidth()
		{
			LogWrite("  proc GetTextWidth書き換え\n");
			uint offset = posGetTextWidthStart;
			PatchByte(data, offset, 0x83);
			offset++;
			PatchByte(data, offset, 0x3D);
			offset++;
			PatchLong(data, offset, addrVarGetTextWidthAddress,
			          string.Format("%03 ${0:X8} varGetTextWidthAddress", addrVarGetTextWidthAddress));
			offset += 4;
			PatchByte(data, offset, 0x00);
			offset++;
			PatchByte(data, offset, 0x75);
			offset++;
			PatchByte(data, offset, 0x29);
			offset++;
			PatchByte(data, offset, 0x68);
			offset++;
//			PatchLong(data, offset, addressBase + posWinMmDll, string.Format("%06 ${0:X8} WINMM.dll", posWinMmDll));
			PatchLong(data, offset, addrWinMmDll, string.Format("%06 ${0:X8} WINMM.dll", addrWinMmDll));
			offset += 4;
			PatchByte(data, offset, 0xFF);
			offset++;
			PatchByte(data, offset, 0x15);
			offset++;
			PatchLong(data, offset, addrGetModuleHandleA,
			          string.Format("%04 ${0:X8} GetModuleHandleA", addrGetModuleHandleA));
			offset += 4;
			PatchByte(data, offset, 0x85);
			offset++;
			PatchByte(data, offset, 0xC0);
			offset++;
			PatchByte(data, offset, 0x75);
			offset++;
			PatchByte(data, offset, 0x01);
			offset++;
			PatchByte(data, offset, 0xF4);
			offset++;
			PatchByte(data, offset, 0x68);
			offset++;
			PatchLong(data, offset, addrGetTextWidth, string.Format("%01 ${0:X8} GetTextWidth", addrGetTextWidth));
			offset += 4;
			PatchByte(data, offset, 0x50);
			offset++;
			PatchByte(data, offset, 0xFF);
			offset++;
			PatchByte(data, offset, 0x15);
			offset++;
			PatchLong(data, offset, addrGetProcAddress, string.Format("%05 ${0:X8} GetProcAddress", addrGetProcAddress));
			offset += 4;
			PatchByte(data, offset, 0x85);
			offset++;
			PatchByte(data, offset, 0xC0);
			offset++;
			PatchByte(data, offset, 0x75);
			offset++;
			PatchByte(data, offset, 0x01);
			offset++;
			PatchByte(data, offset, 0xF4);
			offset++;
			PatchByte(data, offset, 0xA3);
			offset++;
			PatchLong(data, offset, addrVarGetTextWidthAddress,
			          string.Format("%03 ${0:X8} varGetTextWidthAddress", addrVarGetTextWidthAddress));
			offset += 4;
			PatchByte(data, offset, 0x8B);
			offset++;
			PatchByte(data, offset, 0x4D);
			offset++;
			PatchByte(data, offset, 0xEC);
			offset++;
			PatchByte(data, offset, 0x8B);
			offset++;
			PatchByte(data, offset, 0x51);
			offset++;
			PatchByte(data, offset, 0x0C);
			offset++;
			PatchByte(data, offset, 0x81);
			offset++;
			PatchByte(data, offset, 0x7A);
			offset++;
			PatchByte(data, offset, 0x19);
			offset++;
			PatchByte(data, offset, 0x62);
			offset++;
			PatchByte(data, offset, 0x6F);
			offset++;
			PatchByte(data, offset, 0x6C);
			offset++;
			PatchByte(data, offset, 0x64);
			offset++;
			PatchByte(data, offset, 0xB8);
			offset++;
			PatchByte(data, offset, 0x01);
			offset++;
			PatchByte(data, offset, 0x00);
			offset++;
			PatchByte(data, offset, 0x00);
			offset++;
			PatchByte(data, offset, 0x00);
			offset++;
			PatchByte(data, offset, 0x74);
			offset++;
			PatchByte(data, offset, 0x02);
			offset++;
			PatchByte(data, offset, 0x33);
			offset++;
			PatchByte(data, offset, 0xC0);
			offset++;
			PatchByte(data, offset, 0x8B);
			offset++;
			PatchByte(data, offset, 0xD0);
			offset++;
			PatchByte(data, offset, 0x80);
			offset++;
			PatchByte(data, offset, 0xB9);
			offset++;
			PatchByte(data, offset, 0x14);
			offset++;
			PatchByte(data, offset, 0x04);
			offset++;
			PatchByte(data, offset, 0x00);
			offset++;
			PatchByte(data, offset, 0x00);
			offset++;
			PatchByte(data, offset, 0x00);
			offset++;
			PatchByte(data, offset, 0x74);
			offset++;
			PatchByte(data, offset, 0x03);
			offset++;
			PatchByte(data, offset, 0x83);
			offset++;
			PatchByte(data, offset, 0xC8);
			offset++;
			PatchByte(data, offset, 0x02);
			offset++;
			PatchByte(data, offset, 0x50);
			offset++;
			PatchByte(data, offset, 0x8B);
			offset++;
			PatchByte(data, offset, 0x41);
			offset++;
			PatchByte(data, offset, 0x10);
			offset++;
			PatchByte(data, offset, 0x48);
			offset++;
			PatchByte(data, offset, 0x48);
			offset++;
			PatchByte(data, offset, 0x48);
			offset++;
			PatchByte(data, offset, 0x2B);
			offset++;
			PatchByte(data, offset, 0xC2);
			offset++;
			PatchByte(data, offset, 0x8B);
			offset++;
			PatchByte(data, offset, 0x51);
			offset++;
			PatchByte(data, offset, 0x0C);
			offset++;
			PatchByte(data, offset, 0x81);
			offset++;
			PatchByte(data, offset, 0x7A);
			offset++;
			PatchByte(data, offset, 0x0A);
			offset++;
			PatchByte(data, offset, 0x61);
			offset++;
			PatchByte(data, offset, 0x72);
			offset++;
			PatchByte(data, offset, 0x69);
			offset++;
			PatchByte(data, offset, 0x61);
			offset++;
			PatchByte(data, offset, 0x75);
			offset++;
			PatchByte(data, offset, 0x01);
			offset++;
			PatchByte(data, offset, 0x48);
			offset++;
			PatchByte(data, offset, 0x50);
			offset++;
			PatchByte(data, offset, 0xFF);
			offset++;
			PatchByte(data, offset, 0x75);
			offset++;
			PatchByte(data, offset, 0x08);
			offset++;
			PatchByte(data, offset, 0xFF);
			offset++;
			PatchByte(data, offset, 0x15);
			offset++;
			PatchLong(data, offset, addrVarGetTextWidthAddress,
			          string.Format("%03 ${0:X8} varGetTextWidthAddress", addrVarGetTextWidthAddress));
			offset += 4;
			PatchByte(data, offset, 0xE9);
			offset++;
			PatchLong(data, offset, posGetTextWidthEnd - offset - 1,
			          string.Format("%10 ${0:X8} GetTextWidthEnd", posGetTextWidthEnd));
			offset += 4;
			PatchByte(data, offset, 0x90);
			offset++;
			PatchByte(data, offset, 0x90);
			offset++;
			PatchByte(data, offset, 0x90);
			offset++;
			PatchByte(data, offset, 0x90);
			LogWrite("\n");
		}

		private void PatchWinMmDll()
		{
			LogWrite("  WINMM.dll書き換え\n");
			LogWrite(string.Format("  ${0:X8}\n\n", posWinMmDllDef));
			data[posWinMmDllDef] = (byte) '_';
		}

		private void EmbedTextOutDc0()
		{
			uint offset = posTextOutDc0;
			LogWrite("  TextOutDC0埋め込み\n");
			LogWrite(string.Format("  ${0:X8}\n\n", offset));
			data[offset++] = (byte) 'T';
			data[offset++] = (byte) 'e';
			data[offset++] = (byte) 'x';
			data[offset++] = (byte) 't';
			data[offset++] = (byte) 'O';
			data[offset++] = (byte) 'u';
			data[offset++] = (byte) 't';
			data[offset++] = (byte) 'D';
			data[offset++] = (byte) 'C';
			data[offset] = (byte) '0';
		}

		private void EmbedGetTextWidth()
		{
			uint offset = posGetTextWidth;
			LogWrite("  GetTextWidth埋め込み\n");
			LogWrite(string.Format("  ${0:X8}\n\n", offset));
			data[offset++] = (byte) 'G';
			data[offset++] = (byte) 'e';
			data[offset++] = (byte) 't';
			data[offset++] = (byte) 'T';
			data[offset++] = (byte) 'e';
			data[offset++] = (byte) 'x';
			data[offset++] = (byte) 't';
			data[offset++] = (byte) 'W';
			data[offset++] = (byte) 'i';
			data[offset++] = (byte) 'd';
			data[offset++] = (byte) 't';
			data[offset] = (byte) 'h';
		}

		private void EmbedWinMmDll()
		{
			uint offset = posWinMmDll;
			LogWrite("  _INMM.dll埋め込み\n");
			LogWrite(string.Format("  ${0:X8}\n\n", offset));
			data[offset++] = (byte) '_';
			data[offset++] = (byte) 'I';
			data[offset++] = (byte) 'N';
			data[offset++] = (byte) 'M';
			data[offset++] = (byte) 'M';
			data[offset++] = (byte) '.';
			data[offset++] = (byte) 'd';
			data[offset++] = (byte) 'l';
			data[offset] = (byte) 'l';
		}

		private void PatchLatinToUpper()
		{
			LogWrite("  proc LatinToUpper書き換え\n");
			uint offset = posLatinToUpper;
			PatchByte(data, offset, 0xEB);
			offset++;
			PatchByte(data, offset, 0x14);
			LogWrite("\n");
		}

		private List<uint> BinaryScan(byte[] target, byte[] pattern, uint start, uint size)
		{
			StringBuilder sb = new StringBuilder("  Binary:");
			for (int i = 0; i < pattern.Length; i++)
			{
				sb.AppendFormat(" {0:X2}", pattern[i]);
			}
			sb.AppendLine();
			LogWrite(sb.ToString());
			LogWrite(string.Format("  ファイル上の開始位置: ${0:X8}\n", start));
			LogWrite(string.Format("  ファイル上の終了位置: ${0:X8}\n", start + size - 1));

			LogWrite(string.Format("  検索範囲 ${0:X8}～${1:X8} (のこり{2}Bytes)\n", start, start + size - 1, size));
			List<uint> result = new List<uint>();
			for (uint offset = start; offset <= start + size - pattern.Length; offset++)
			{
				if (IsBinaryMatch(target, pattern, offset))
				{
					result.Add(offset);
					LogWrite(string.Format("  *** Find ${0:X8}\n", offset));
					LogWrite(string.Format("  検索範囲 ${0:X8}～${1:X8} (のこり{2}Bytes)\n", offset + 1, start + size - 1,
					                       start + size - offset - 1));
				}
			}
			return result;
		}

		private static bool IsBinaryMatch(byte[] target, byte[] pattern, uint offset)
		{
			int i;
			for (i = 0; i < pattern.Length; i++)
			{
				if (target[offset + i] != pattern[i])
					return false;
			}
			return true;
		}

		private void PatchByte(byte[] target, uint offset, byte val)
		{
			LogWrite(string.Format("  ${0:X8} {1:X2}->{2:X2}\n", offset, target[offset], val));
			target[offset] = val;
		}

		private void PatchLong(byte[] target, uint offset, uint val, string message)
		{
			LogWrite(string.Format("  ${0:X8} {1:X8}->{2:X8} ({3})\n", offset, GetLong(offset), val, message));
			target[offset] = (byte) (val & 0x000000FF);
			target[offset + 1] = (byte) ((val & 0x0000FF00) >> 8);
			target[offset + 2] = (byte) ((val & 0x00FF0000) >> 16);
			target[offset + 3] = (byte) ((val & 0xFF000000) >> 24);
		}

		private ushort GetWord(uint offset)
		{
			return (ushort) (data[offset] + data[offset + 1]*0x0100);
		}

		private uint GetLong(uint offset)
		{
			return
				(uint) (data[offset] + data[offset + 1]*0x00000100 + data[offset + 2]*0x00010000 + data[offset + 3]*0x01000000);
		}

		private string GetString(uint offset, int max)
		{
			string result = "";
			for (int i = 0;; i++)
			{
				if (max > 0 && i >= max)
				{
					break;
				}
				if (data[offset + i] == 0x00)
				{
					break;
				}
				result += (char) data[offset + i];
			}
			return result;
		}

		private static uint Ceiling(uint val, uint unit)
		{
			uint result = (val + unit - 1)/unit;
			result *= unit;
			return result;
		}

		private void LogWrite(string s)
		{
			mainForm.AppendLog(s);
		}
	}
}